﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;

using ESRI.ArcGISExplorer;
using ESRI.ArcGISExplorer.Application;
using ESRI.ArcGISExplorer.Mapping;
using ESRI.ArcGISExplorer.Geometry;
using ESRI.ArcGISExplorer.Data;
using ESRI.ArcGISExplorer.Threading;

using gp = TerrainProfile900.ElevationGPServer;

namespace TerrainProfile900
{
    public class Button : ESRI.ArcGISExplorer.Application.Button
    {
        // MAP DISPLAY
        private MapDisplay _md = null;
        // BACKGROUND WORKER
        private ESRI.ArcGISExplorer.Threading.BackgroundWorker _bgWorker;
        // PROFILE NOTE
        private Note _profileNote;
        // PROGRESS FORM
        private ProgressForm _progressForm;

        // CONSTRUCTOR
        public Button()
        {
            // GET THE ACTIVE MAP
            _md = ESRI.ArcGISExplorer.Application.Application.ActiveMapDisplay;

            // CREATE BACKGROUND WORKER AND LISTEN TO EVENTS
            _bgWorker = new ESRI.ArcGISExplorer.Threading.BackgroundWorker();
            _bgWorker.DoWork += new ESRI.ArcGISExplorer.Threading.DoWorkEventHandler(_bgWorker_DoWork);
            _bgWorker.ProgressChanged += new ESRI.ArcGISExplorer.Threading.ProgressChangedEventHandler(_bgWorker_ProgressChanged);
            _bgWorker.RunWorkerCompleted += new ESRI.ArcGISExplorer.Threading.RunWorkerCompletedEventHandler(_bgWorker_RunWorkerCompleted);            
        }

        // ON UPDATE
        public override void OnUpdate()
        {
            // DISABLE IF CURRENTLY BUSY
            this.Enabled = (!_md.IsTracking) && (!_bgWorker.IsBusy);
            base.OnUpdate();
        }
              
        // USER CLICKS TOOL
        public override void OnClick()
        {
            // TRACK USER POLYLINE
            Polyline trackedPath = _md.TrackPolyline(Color.Gold, 4.0d);
            if (trackedPath == null)
                return;

            // PROFILE NOTE SYMBOL
            Symbol profileTrackSym = Symbol.CreateLine(Color.Gold);
            profileTrackSym.Size = 4d;

            // CREATE NOTE AND ADD TO CONTENTS
            _profileNote = new Note("Terrain Profile", trackedPath, profileTrackSym);
            _profileNote.Graphic.Placement3D = Placement3D.AttachToSurface;
            _profileNote.Popup.Size = new Size(655, 310);
            _md.Map.ChildItems.AddFirst(_profileNote);

            // PROGRESS DIALOG
            _progressForm = new ProgressForm(_bgWorker);
            
            // RUN ANALYSIS IN BACKGROUND WORKER
            _bgWorker.RunWorkerAsync(trackedPath.ToXmlString());
            
            // SHOW PROGRESS DIALOG
            _progressForm.ShowDialog(Application.Window);
        }       
        
        /// <summary>
        /// Called on the worker thread.
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void _bgWorker_DoWork(object sender, ESRI.ArcGISExplorer.Threading.DoWorkEventArgs e)
        {
            // GET BACKGROUND WORKER
            ESRI.ArcGISExplorer.Threading.BackgroundWorker bgWorker = (ESRI.ArcGISExplorer.Threading.BackgroundWorker)sender;

            // UPDATE INITIAL PROGRESS
            int count = 0;
            bgWorker.ReportProgress(++count, "Calculating Profile...");

            // Polyline geometry is passed into background thread as xml string argument.
            Polyline userPath = Polyline.CreateFromXmlString(Convert.ToString(e.Argument));

            // IMAGE SIZE
            int imageWidth = 600;
            int imageHeight = 200;

            // CANCELLATION
            if (bgWorker.CancellationPending)
                return;

            // GET ELEVATION SERVICE
            gp.ESRI_Elevation_World_GPServer elevationService = new gp.ESRI_Elevation_World_GPServer();

            // CANCELLATION
            if (bgWorker.CancellationPending)
                return;

            // GET TOOL INFO
            gp.GPToolInfo gpti = elevationService.GetToolInfo("ProfileService");
            // GET PARAMETER INFO
            gp.GPParameterInfo[] paramInfo = gpti.ParameterInfo;

            // CANCELLATION
            if (bgWorker.CancellationPending)
                return;

            // GET FEATURERECORDSETLAYER
            gp.GPFeatureRecordSetLayer recordSetLayer = (gp.GPFeatureRecordSetLayer)paramInfo[0].Value;

            // CREATE PARAMETER VALUES
            gp.GPLong gpImageWidth = new gp.GPLong();
            gpImageWidth.Value = imageWidth;
            gp.GPLong gpImageHeight = new gp.GPLong();
            gpImageHeight.Value = imageHeight;
            gp.GPBoolean gpDisplaySegments = new gp.GPBoolean();
            gpDisplaySegments.Value = true;

            // CANCELLATION
            if (bgWorker.CancellationPending)
                return;

            // CONVER E3 POLYLINE TO SOAP POLYLINEN
            gp.PolylineN userPathN = (gp.PolylineN)SoapConverter.GeometryToSoap<Polyline, gp.PolylineN>(userPath);
            userPathN.HasZ = false;

            // SET OUTPUT RESULT OPTIONS
            gp.GPResultOptions resultOptions = new gp.GPResultOptions();
            resultOptions.SpatialReference = userPathN.SpatialReference;

            // CANCELLATION
            if (bgWorker.CancellationPending)
                return;

            // CREATE ARRAY OF VALUES
            object[] vals = new object[3];
            vals[0] = 1;
            vals[1] = userPathN;
            vals[2] = 0.0;
            // CREATE FEATURE
            gp.Record feature = new gp.Record();
            // SET FEATURE VALUES
            feature.Values = vals;
            // CREATE ARRAY OF FEATURES
            gp.Record[] features = new gp.Record[1];
            // ADD FEATURE TO ARRAY OF FEATURES
            features[0] = feature;
            // SET ARRAY OF FEATURES AS RECORDS IN THE INPUT RECORDSET
            recordSetLayer.RecordSet.Records = features;

            // SET INPUT PARAMETERS
            gp.GPValue[] gpInputValues = new gp.GPValue[4];
            gpInputValues[0] = recordSetLayer;
            gpInputValues[1] = gpImageWidth;
            gpInputValues[2] = gpImageHeight;
            gpInputValues[3] = gpDisplaySegments;

            // CANCELLATION
            if (bgWorker.CancellationPending)
                return;

            // SUBMINT JOB TO GPSERVICE                        
            string jobID = elevationService.SubmitJob(gpti.Name, gpInputValues, resultOptions, null);

            // CHECK JOB STATUS AND CONTINUE UNTIL FINISHED OR ERROR                        
            gp.esriJobStatus jobStatus;
            gp.JobMessage[] jobMessages;
            // MESSAGE DESCRIPTION
            string satusMsg = string.Empty;

            do
            {   
                // GET JOB STATUS
                jobStatus = elevationService.GetJobStatus(jobID);
                satusMsg = jobStatus.ToString();
                switch (jobStatus)
                {
                    // CHECK FOR PROBLEMS
                    case gp.esriJobStatus.esriJobFailed:
                    case gp.esriJobStatus.esriJobTimedOut:
                    case gp.esriJobStatus.esriJobCancelled:   
                        //  THROW EXCEPTION
                        throw new Exception(satusMsg);                        
                    default:
                        break;
                }

                // GET MESSAGES
                jobMessages = elevationService.GetJobMessages(jobID);
                if (jobMessages.Length > 1)
                {
                    // THE LAST MESSAGE IS NORMALLY ONE WITH THE START TIME OF THE
                    // CURRENT STEP BEING EXECUTED, SO WE'LL USE THE MESSAGE BEFORE
                    // THE LAST ONE TO GET THE NAME OF THE STEP BEING EXECUTED...
                    satusMsg = jobMessages[jobMessages.Length - 2].MessageDesc;
                    if (satusMsg.Contains("Time")) // WE SOMETIMES STILL GET A 'TIME' MESSAGE...
                        satusMsg = jobMessages[jobMessages.Length - 1].MessageDesc;

                    satusMsg = satusMsg.Split(':')[0];
                }

                // UPDATE STATUS                                                
                bgWorker.ReportProgress(++count, satusMsg);

                // PAUSE BEFORE CHECKING STATUS AGAIN
                System.Threading.Thread.Sleep(500);

                // EXIT LOOP IF WE HAVE SUCCEEDED OR IF CANCELLED             
            } while ((jobStatus != gp.esriJobStatus.esriJobSucceeded)  && (!bgWorker.CancellationPending));

            // CANCELLATION
            if (bgWorker.CancellationPending)
                return;

            // GET GP JOB RESULT                
            gp.GPResult result = elevationService.GetJobResult(jobID, null, resultOptions);

            // GET FIRST FEATURE FROM FEATURESET
            gp.GPFeatureRecordSetLayer gpOutputRS = (gp.GPFeatureRecordSetLayer)result.Values[0];
            gp.RecordSet recordSet = gpOutputRS.RecordSet;
            gp.Record resultFeature = recordSet.Records[0];

            // GET PROFILE IMAGE URL
            string profileImageURL = resultFeature.Values[6].ToString();

            // SET PROFILE RESULT                
            e.Result = profileImageURL;
        }

        /// <summary>
        /// Called on the UI thread.
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void _bgWorker_ProgressChanged(object sender, ESRI.ArcGISExplorer.Threading.ProgressChangedEventArgs e)
        {
            // UPDATE PROFILE MESSAGES
            _progressForm.StatusMessage = string.Format("Terrain Profile: {0}", e.Message);
        }

        /// <summary>
        /// Called on the UI thread.
        /// 
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void _bgWorker_RunWorkerCompleted(object sender, ESRI.ArcGISExplorer.Threading.RunWorkerCompletedEventArgs e)
        {
            // CLOSE PROGRESS DIALOG
            _progressForm.Close();

            // CHECK STATUS
            if (e.Cancelled) // WAS THE PROCESS CANCELLED
            {
                _md.Map.ChildItems.Remove(_profileNote);
                return;
            }
            else if (e.Error != null) // WAS THERE AN ERROR THROWN IN BACKGROUND WORKER
            {
                // PROFILE ERROR SYMBOL
                Symbol profileErrorSym = Symbol.Line.Solid.Red;
                profileErrorSym.Size = 4d;

                // UPDATE PROFILE NOTE
                _profileNote.Name = string.Format("Terrain Profile Error: {0}", e.Error.Message);
                _profileNote.Graphic.Symbol = profileErrorSym;
                _profileNote.Popup.Content = e.Error.ToString();
            }
            else
            {
                // PROFILE IMAGE URL
                string profileImageURL = e.Result.ToString();

                // PROFILE SYMBOL
                Symbol profileSym = Symbol.CreateLine(Color.Blue);
                profileSym.Size = 4d;

                // UPDATE PROFILE NOTE
                _profileNote.Name = "Terrain Profile";
                _profileNote.Graphic.Symbol = profileSym;
                _profileNote.Popup.Content = string.Format("<html><img src='{0}' width='100%' height='100%' /><html>", profileImageURL);
            }

            // SHOW INFOPOPUP WINDOW            
            _profileNote.Popup.Activate();
        }
    }
}

